"""
Message Module
"""

from email import message_from_bytes
from email.header import decode_header
from email.utils import parsedate_to_datetime

from email_profile.data import DataClass
from email_profile.models import AttachmentModel, EmailModel


class Message:

    def __init__(self, message: bytes, id: int) -> None:
        self.message = message_from_bytes(message)
        self.id = id
        self.body_text_plain = ""
        self.body_text_html = ""
        self.data = DataClass()

    def decode_field(self, header) -> str:
        field = ""
        for sub in decode_header(header):
            encoding = sub[1] or "utf-8"
            try:
                field += sub[0].decode(encoding)
            except Exception:
                field += sub[0]
        return field

    def parsedate_to_datetime(self, date):
        try:
            return parsedate_to_datetime(date)
        except ValueError:
            pass
        return None

    def get_content(self, part) -> None:
        content_type = part.get_content_type()

        if content_type == "text/plain":
            try:
                self.body_text_plain = part.get_payload(decode=True).decode()
            except Exception:
                self.body_text_plain = part.get_payload(decode=True)

        if content_type == "text/html":
            try:
                self.body_text_html = part.get_payload(decode=True).decode()
            except Exception:
                self.body_text_html = part.get_payload(decode=True)
                if isinstance(self.body_text_html, bytes):
                    self.body_text_html = part.get_payload()

        if "attachment" in str(part.get("Content-Disposition")):
            filename = part.get_filename()
            if filename:
                model = AttachmentModel(
                    id=self.id,
                    file_name=filename,
                    content_type=part.get_content_type(),
                    content_ascii=part.get_payload().encode("ascii")
                )
                self.data.add_attachment(model)

    def result(self) -> EmailModel:
        for part in self.message.walk():
            self.get_content(part=part)

        model = EmailModel(
            id=self.id,
            body_text_plain=self.body_text_plain,
            body_text_html=self.body_text_html,
            return_path=self.message.get("Return-Path"),
            delivered_to=self.message.get("Delivered-To"),
            received=self.message.get("Received"),
            dkim_signature=self.message.get("DKIM-Signature"),
            content_type=self.message.get_content_type(),
            date=self.parsedate_to_datetime(self.message.get("Date")),
            from_who=self.decode_field(self.message.get("From")),
            mime_version=self.message.get("Mime-Version"),
            message_id=self.message.get("Message-ID"),
            subject=self.decode_field(self.message["Subject"]),
            reply_to=self.message.get("Reply-To"),
            precedence=self.message.get("Precedence"),
            x_sg_eid=self.message.get("X-SG-EID"),
            x_sg_id=self.message.get("X-SG-ID"),
            to_who=self.message.get("To"),
            x_entity_id=self.message.get("X-Entity-ID")
        )
        self.data.add_email(model)

        return self.data



from json import loads
from pathlib import Path

from email_profile.data import DataClass
from email_profile.models import AttachmentModel, EmailModel

from tests.conftest import text_html   # noqa: F401


def test_instance_data():
    data = DataClass()

    assert isinstance(data, DataClass)
    assert data.email is None
    assert data.attachments == list()


def test_data_add_email():
    email = EmailModel()
    data = DataClass()
    data.add_email(model=email)

    assert isinstance(data.email, EmailModel)
    assert data.email == email


def test_data_add_attachment():
    attachment = AttachmentModel()
    data = DataClass()
    data.add_attachment(model=attachment)

    assert len(data.attachments) == 1
    assert isinstance(data.attachments[0], AttachmentModel)
    assert data.attachments[0] == attachment


def test_data_json():
    email = EmailModel(id=1)

    data = DataClass()
    data.add_email(model=email)

    response = data.json(create_file=False)

    assert isinstance(response, dict)
    assert response.get("id") == 1


def test_data_json_create_file():
    file_name = "42.json"
    path = Path("tests", "json")

    email = EmailModel(id=42)

    data = DataClass()
    data.add_email(model=email)

    data.json(
        path=path,
        create_file=True
    )

    assert path.joinpath(file_name).exists()

    with open(path.joinpath(file_name), mode="r") as file:
        email = loads(file.read())
        assert email.get("id") == 42

    path.joinpath(file_name).unlink()
    path.rmdir()


def test_data_html(text_html):  # noqa: F811
    email = EmailModel(
        id=42,
        body_text_html=text_html
    )

    data = DataClass()
    data.add_email(model=email)

    response = data.html(create_file=False)

    assert "Email Profile" in response


def test_data_html_create_file(text_html):  # noqa: F811
    file_name = "index.html"
    path = Path("tests", "html")

    email = EmailModel(
        id=42,
        body_text_html=text_html
    )

    data = DataClass()
    data.add_email(model=email)

    data.html(
        path=path,
        create_file=True
    )

    assert path.joinpath("42", file_name).exists()

    with open(path.joinpath("42", file_name), mode="r") as file:
        email = file.read()
        assert "Email Profile" in email

    path.joinpath("42", file_name).unlink()
    path.joinpath("42").rmdir()
    path.rmdir()



from dataclasses import dataclass, field

from email_profile.abstract import AbstractModel


@dataclass
class AttachmentModel(AbstractModel):

    class Meta:
        table_name = 'attachment'

    id: int = field(default=None)
    file_name: str = field(default=None)
    content_type: str = field(default=None)
    content_ascii: str = field(default=None)



"""
Model Module
"""

from dataclasses import dataclass
from abc import abstractclassmethod, ABC


class Validations:

    def __post_init__(self):
        for name, _field in self.__dataclass_fields__.items():
            method_name = f"validate_{name}"

            if hasattr(self, method_name):
                if method := getattr(self, method_name):
                    setattr(self, name, method(field=_field))


@dataclass
class AbstractModel(ABC, Validations):

    @abstractclassmethod
    class Meta:
        table_name = None

    def get_table_name(self):
        return self.Meta.table_name

    def get_fields(self):
        fields = self.__dataclass_fields__.items()
        return [name for name, field in fields]

    def get_values(self):
        values = [value for value in self.__dict__.items()]
        return [f"'{str(item)}'" for x, item in values]



"""
Data Module
"""

import json

from pathlib import Path
from typing import List, Dict
from abc import abstractmethod, ABC

from email_profile.utils import mkdir


class DataAbstract(ABC):

    _id = None

    def __init__(self) -> None:
        self.email: object = None
        self.attachments: List[object] = list()

    def add_email(self, model: object):
        self._id = str(model.id)
        self.email = model

    def add_attachment(self, model: object):
        self.attachments.append(model)

    @abstractmethod
    def json(self) -> None:
        pass

    @abstractmethod
    def html(self) -> None:
        pass


class DataClass(DataAbstract):

    def json(self,
             path: str = "json",
             create_file: bool = True) -> Dict:
        data = self.email.__dict__

        if create_file:
            path = Path(path)
            mkdir(path)

            with open(
                file=path.joinpath(f"{self._id}.json"),
                mode="w",
                errors="ignore"
            ) as file:
                file.write(
                    json.dumps(data, indent=4)
                )

        return data

    def html(self,
             path: str = "html",
             create_file: bool = True) -> str:
        data = self.email.body_text_html

        if create_file:
            path = Path(path, self._id)
            mkdir(path)

            with open(
                file=path.joinpath("index.html"),
                mode="w",
                errors="ignore"
            ) as file:
                file.write(
                    self.email.body_text_html
                )

        return data



from datetime import datetime
from dataclasses import dataclass, field

from email_profile.abstract import AbstractModel


@dataclass
class EmailModel(AbstractModel):

    class Meta:
        table_name = 'email'

    id: int = field(default=None)
    body_text_plain: str = field(default=None)
    body_text_html: str = field(default=None)
    return_path: str = field(default=None)
    delivered_to: str = field(default=None)
    received: str = field(default=None)
    dkim_signature: str = field(default=None)
    received: str = field(default=None)
    content_type: str = field(default=None)
    date: datetime = field(default=None)
    from_who: str = field(default=None)
    mime_version: str = field(default=None)
    message_id: str = field(default=None)
    subject: str = field(default=None)
    reply_to: str = field(default=None)
    precedence: str = field(default=None)
    x_sg_eid: str = field(default=None)
    x_sg_id: str = field(default=None)
    to_who: str = field(default=None)
    x_entity_id: str = field(default=None)

    def validate_body_text_plain(self, field) -> str:
        if self.body_text_plain:
            return str(self.body_text_plain)
        return ''

    def validate_date(self, field) -> str:
        if self.date:
            return self.date.strftime('%Y-%d-%m %H:%M:%S')
        return ''



from dataclasses import dataclass, field

from email_profile.abstract import AbstractModel


@dataclass
class MailBoxModel(AbstractModel):

    class Meta:
        table_name = 'mailbox'

    id: int = field(default=None)
    name: str = field(default=None)



"""
Staus Module
"""

from typing import List


class StatusResponse:

    def __init__(self, type: bool, message: str) -> None:
        self.type = type
        self.message = message


class Status:

    OK = 'OK'
    NO = 'NO'
    BAD = 'BAD'

    @staticmethod
    def validate_status(
            status: str,
            raise_error: bool = True) -> StatusResponse:
        data = {
            Status.OK: (
                True,
                "Login completed, now in authenticated state"),
            Status.NO: (
                False,
                "Login failure: user name or password rejected"),
            Status.BAD: (
                False,
                "Command unknown or arguments invalid")
        }.get(status, (False, "Unknown error"))

        if raise_error:
            if not data[0]:
                raise Exception(data[1])

        return StatusResponse(
            type=data[0],
            message=data[1]
        )

    @staticmethod
    def validate_data(data: List[str]) -> List[str]:
        data = [x.decode() for x in data[0].split()]
        if len(data) == 1 and not data[0]:
            return []
        return data



"""
Controller Module
"""

from abc import abstractmethod, ABC


class AbstractController(ABC):

    def __init__(self, model=None, data=None) -> None:
        self.model = model
        self.data = data

    @abstractmethod
    def create(self) -> None:
        pass

    @abstractmethod
    def read(self) -> None:
        pass

    @abstractmethod
    def update(self) -> None:
        pass

    @abstractmethod
    def delete(self) -> None:
        pass
